import sqlparse

from metrics_layer.core.convert import MQLConverter
from metrics_layer.core.exceptions import MetricsLayerException, QueryError
from metrics_layer.core.parse import ProjectLoader
from metrics_layer.core.sql import SQLQueryResolver
from metrics_layer.core.sql.arbitrary_merge_resolve import ArbitraryMergedQueryResolver
from metrics_layer.core.sql.query_errors import ParseError


class DBConnectionError(MetricsLayerException):
    pass


class MetricsLayerConnection:
    def __init__(
        self,
        location: str = None,
        branch: str = "master",
        project=None,
        connections: list = [],
        user: dict = None,
        **kwargs,
    ):
        self.location, self.branch, self._raw_connections = location, branch, connections
        self.kwargs = kwargs
        self._user = user
        self.branch_options = None
        self._project = None
        if project is not None:
            self._project_passed = True
            self._project = project
            self._project.set_user(self._user)
            self.branch_options = []

    def set_user(self, user: dict):
        self._user = user
        self.project.set_user(self._user)

    def load(self, private_key: str = None):
        if self.location is not None:
            self._loader = ProjectLoader(self.location, self.branch, self._raw_connections)
            self._project = self._loader.load(private_key=private_key)
            self._project.set_user(self._user)
            self.branch_options = self._loader.get_branch_options()
        elif self._project_passed:
            # Project is passed in explicitly, nothing else to do
            pass
        else:
            raise DBConnectionError(
                "No project or location specified. Pass either the location "
                "(a path or a github url) or a project object."
            )

    @property
    def profiles_path(self):
        return ProjectLoader.profiles_path()

    @property
    def project(self):
        if self._project is None:
            raise QueryError("You must call the load() method before accessing the project.")
        return self._project

    def get_branch_options(self):
        if self.branch_options is None:
            raise QueryError("You must call the load() method before accessing the branch options.")
        return self.branch_options

    def add_connections(self, additional_raw_connections: list):
        self._raw_connections = self._raw_connections + additional_raw_connections

    @property
    def connections(self):
        self._connections = ProjectLoader.load_connections(self._raw_connections)
        return self._connections

    def list_connections(self, names_only=False):
        connections = self.connections
        if names_only:
            return [c.name for c in connections]
        return connections

    def get_connection(self, connection_name: str):
        return next((c for c in self.list_connections() if c.name == connection_name), None)

    def query(
        self,
        metrics: list = [],
        dimensions: list = [],
        funnel: dict = {},
        where: list = [],
        having: list = [],
        order_by: list = [],
        sql: str = None,
        **kwargs,
    ):
        query, connection = self.get_sql_query(
            sql=sql,
            metrics=metrics,
            dimensions=dimensions,
            funnel=funnel,
            where=where,
            having=having,
            order_by=order_by,
            **{**self.kwargs, **kwargs},
            return_connection=True,
        )
        df = self.run_query(query, connection, **kwargs)
        return df

    def get_sql_query(
        self,
        metrics: list = [],
        dimensions: list = [],
        funnel: dict = {},
        where: list = [],
        having: list = [],
        order_by: list = [],
        sql: str = None,
        merged_queries: list = [],
        **kwargs,
    ):
        if sql:
            converter = MQLConverter(
                sql, project=self.project, connections=self.connections, **{**self.kwargs, **kwargs}
            )
            connection = converter.connection
            query = converter.get_query()
        elif metrics or dimensions:
            resolver = SQLQueryResolver(
                metrics=metrics,
                dimensions=dimensions,
                funnel=funnel,
                where=where,
                having=having,
                order_by=order_by,
                project=self.project,
                connections=self.connections,
                **{**self.kwargs, **kwargs},
            )
            connection = resolver.connection
            query = resolver.get_query()
        elif len(merged_queries) > 0:
            # This kwarg is meaningless in the context of the merged query resolver
            # But it can mess up sub queries if it's not popped here
            kwargs.pop("merged_result", None)
            resolver = ArbitraryMergedQueryResolver(
                merged_queries=merged_queries,
                where=where,
                having=having,
                order_by=order_by,
                project=self.project,
                connections=self.connections,
                **{**self.kwargs, **kwargs},
            )
            connection = resolver.connection
            query = resolver.get_query()
        else:
            raise QueryError(
                'No metrics or dimensions specified. Please provide either "metrics" or "dimensions"'
            )

        if kwargs.get("pretty", False):
            query = self.pretty_sql(query)

        if kwargs.get("return_connection", False):
            return query, connection

        if kwargs.get("return_query_kind", False):
            return query, resolver.query_kind
        return query

    def list_fields(self, view_name: str = None, names_only: bool = False, show_hidden: bool = False):
        all_fields = self.project.fields(view_name=view_name, show_hidden=show_hidden)
        if names_only:
            return [m.name for m in all_fields]
        return all_fields

    def get_field(self, field_name: str, view_name: str = None):
        models = self.list_models()
        kws = {}
        if len(models) == 1:
            kws = {"model_name": models[0].name}
        return self.project.get_field(field_name, view_name=view_name, **kws)

    def list_metrics(self, view_name: str = None, names_only: bool = False, show_hidden: bool = False):
        all_fields = self.project.fields(view_name=view_name, show_hidden=show_hidden)
        metrics = [f for f in all_fields if f.field_type == "measure"]
        if names_only:
            return [m.name for m in metrics]
        return metrics

    def get_metric(self, metric_name: str, view_name: str = None):
        metrics = self.list_metrics(view_name=view_name)
        try:
            metric = next((m for m in metrics if m.equal(metric_name)))
            return metric
        except ParseError as e:
            raise e(f"Could not find metric {metric_name} in the project config")

    def list_dimensions(self, view_name: str = None, names_only: bool = False, show_hidden: bool = False):
        all_fields = self.project.fields(view_name=view_name, show_hidden=show_hidden)
        dimensions = [f for f in all_fields if f.field_type in {"dimension", "dimension_group"}]
        if names_only:
            return [d.name for d in dimensions]
        return dimensions

    def get_dimension(self, dimension_name: str, view_name: str = None):
        dimensions = self.list_dimensions(view_name=view_name)
        try:
            dimension = next((d for d in dimensions if d.equal(dimension_name)))
            return dimension
        except ParseError as e:
            raise e(f"Could not find dimension {dimension_name} in the project config")

    def list_views(self, names_only=False):
        views = self.project.views()
        if names_only:
            return [v.name for v in views]
        return views

    def get_view(self, view_name: str):
        return self.project.get_view(view_name)

    def list_models(self, names_only=False):
        models = self.project.models()
        if names_only:
            return [m.name for m in models]
        return models

    def get_model(self, model_name: str):
        return self.project.get_model(model_name)

    def list_dashboards(self, names_only=False):
        dashboards = self.project.dashboards()
        if names_only:
            return [d.name for d in dashboards]
        return dashboards

    def get_dashboard(self, dashboard_name: str):
        return self.project.get_dashboard(dashboard_name)

    def get_all_profiles(self, names_only: bool = False):
        raise NotImplementedError()

    @staticmethod
    def get_connections_from_profile(profile_name: str, target: str = None):
        return ProjectLoader.get_connections_from_profile(profile_name, target)

    @staticmethod
    def pretty_sql(sql: str, keyword_case="lower"):
        return sqlparse.format(sql, reindent=True, keyword_case=keyword_case)



import os

from metrics_layer.core.exceptions import MetricsLayerException
from metrics_layer.core.model import Project
from metrics_layer.core.sql.query_design import MetricsLayerDesign
from metrics_layer.integrations.metricflow.metricflow_to_zenlytic import (
    convert_mf_project_to_zenlytic_project,
    load_mf_project,
)

from .project_reader_base import ProjectReaderBase


class MetricflowParsingException(MetricsLayerException):
    pass


class MetricflowProjectReader(ProjectReaderBase):
    def load(self) -> tuple:
        if self.dbt_project is None:
            raise MetricflowParsingException(
                "No dbt project found. Make sure you have a dbt_project.yml file, and it is accessible to"
                " Zenlytic."
            )

        self.project_name = self.dbt_project["name"]

        if self.zenlytic_project and "metricflow-path" in self.zenlytic_project:
            metricflow_path = self.zenlytic_project["metricflow-path"]
            metricflow_path = os.path.join(self.repo.folder, metricflow_path)
        else:
            metricflow_path = self.dbt_folder

        metricflow_project = load_mf_project(metricflow_path)

        dbt_profile_name = self.dbt_project.get("profile", self.project_name)
        if self.zenlytic_project:
            self.profile_name = self.zenlytic_project.get("profile", dbt_profile_name)
        else:
            self.profile_name = dbt_profile_name

        topic_folders = self.get_folders("topic-paths", raise_errors=False)
        model_folders = self.get_folders("model-paths", raise_errors=False)
        all_folders = model_folders + topic_folders

        file_names = self.search_for_yaml_files(all_folders)

        models = []
        topics = []
        for fn in file_names:
            yaml_dict = self.read_yaml_file(fn)
            yaml_dict["_file_path"] = os.path.relpath(fn, start=self.repo.folder)

            yaml_type = yaml_dict.get("type")
            if yaml_type == "model":
                models.append(yaml_dict)
            elif yaml_type == "topic":
                topics.append(yaml_dict)

        if len(models) == 1:
            model_dict = models[0]
        elif len(models) > 1:
            raise MetricflowParsingException(
                "Multiple models found in model-paths. Only one model is supported with Metricflow."
            )
        else:
            model_dict = {}

        models, views, errors = convert_mf_project_to_zenlytic_project(
            metricflow_project, self.profile_name, self.profile_name, model_dict
        )
        if self.zenlytic_project.get("use_default_topics", True):
            topics += self.derive_topics(models, views, models[0]["name"])

        return models, views, [], topics, errors

    def derive_topics(self, models: list, views: list, model_name: str) -> list:
        project = Project(models, views)
        graph = project.join_graph.graph
        sorted_components = project.join_graph._strongly_connected_components(graph)

        design = MetricsLayerDesign(
            no_group_by=False,
            query_type="SNOWFLAKE",  # this doesn't matter
            field_lookup={},
            model=None,
            project=project,
        )

        topic_views = []
        topics = {}

        for component_set in sorted_components:
            subgraph_nodes = project.join_graph._subgraph_nodes_from_components(graph, component_set)
            design._join_subgraph = project.join_graph.subgraph(subgraph_nodes)
            ordered_components = design.determine_join_order(subgraph_nodes)

            if len(ordered_components) == 0:
                ordered_views = subgraph_nodes
            else:
                ordered_views = [view for components in ordered_components for view in components]
                ordered_views = list(sorted(set(ordered_views), key=lambda x: ordered_views.index(x)))

            topic_views.append(ordered_views)

        # make sure all views have been used in a topic
        all_views = set(v.name for v in project.views())
        all_topic_views = set([topic_view for topic_view_set in topic_views for topic_view in topic_view_set])
        assert all_views == all_topic_views

        # make sure each topic has at least one view
        for tv in topic_views:
            assert len(tv) > 0

        topics = []
        for i, topic_view_list in enumerate(topic_views):
            base_view = topic_view_list[0]
            topic_data = {
                "label": f"{base_view.replace('_', ' ').title()}",
                "base_view": base_view,
                "model_name": model_name,
                "views": {},
            }
            for view in topic_view_list[1:]:
                try:
                    design._join_subgraph = project.join_graph.subgraph([base_view, view])
                    ordered_join_components = design.determine_join_order([base_view, view])
                    if len(ordered_join_components) == 1:
                        join = project.join_graph.get_join(base_view, view)
                    else:
                        connecting_view, destination_view = ordered_join_components[-1]
                        join = project.join_graph.get_join(connecting_view, destination_view)

                    join_type = join._definition.get("type")
                    relationship = join._definition.get("relationship")
                    sql_on = join._definition.get("sql_on")

                    if join_type is not None and relationship is not None and sql_on is not None:
                        topic_data["views"][view] = {
                            "join": {
                                "join_type": join_type,
                                "relationship": relationship,
                                "sql_on": sql_on,
                            }
                        }

                except Exception as e:
                    print(f"error_getting_join_for_topic: {base_view} and {view}: {e}")

            topics.append(topic_data)

        return topics



class WeekStartDayTypes:
    monday = "monday"
    tuesday = "tuesday"
    wednesday = "wednesday"
    thursday = "thursday"
    friday = "friday"
    saturday = "saturday"
    sunday = "sunday"
    default = monday
    options = [monday, tuesday, wednesday, thursday, friday, saturday, sunday]
